/*
 * @Descripttion: move_action.cpp Project
 * @version: Project v0.1
 * @Author: lenotary
 * @Date: 2021/4/21 下午1:24
 * @LastEditor: lenotary
 * @LastEditTime: 2021/4/21 下午1:24
 */

#include "move_action.h"

/**
 * @Title        :
 * @MethodName   :  CMoveAction
 * @param    in  :  ros::NodeHandle &nh
 * @Return       :  null
 * @Exception    :
 * @Description  :  析构函数，初始化订阅
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午1:42
 */

CMoveAction::CMoveAction(ros::NodeHandle &nh) : m_iNh_(nh) {
    m_iMagLineDataSub_ = nh.subscribe("/device/mag_line_data", 1, &CMoveAction::MagLineDataCallback, this);
    m_iVelSub_ = nh.subscribe("/robot_vel", 1, &CMoveAction::VelCallback, this);
    InitParam();
}

/**
 * @Title        :
 * @MethodName   :  MagLineDataCallback
 * @param    in  :  user_msgs::MagLineData::ConstPtr
 * @Return       :  null
 * @Exception    :
 * @Description  :  磁条导航回调函数
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午1:45
 */
void CMoveAction::MagLineDataCallback(const user_msgs::MagLineData::ConstPtr &msg) {
    m_iMagLineData_ = *msg;
}

/**
 * @Title        :
 * @MethodName   :  SonarCallback
 * @param int    :  msg
 * @Return       :
 * @Exception    :
 * @Description  :  订阅超声波数据
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午3:22
 */
void CMoveAction::SonarCallback(const std_msgs::Float32MultiArray::ConstPtr &msg) {
    m_iSonarList_ = *msg;
}

/**
 * @Title        :
 * @MethodName   :  TracingPd
 * @param   in   :  speed (设置速度)
 * @param   out  :  twist （引用设置速度，反馈到调用者）
 * @Return       :  bool
 * @Exception    :
 * @Description  :  PD控制巡线
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午1:46
 */

bool CMoveAction::TracingPd(double speed, geometry_msgs::Twist &twist) {
    static double speedPrev{0};
    static double speedAdd{0};
    static double speedStart{0};
    static double last_error{0};

    if (speed != speedPrev) {
        speedAdd = (speed - m_iVel_.linear.x) / 25;
        speedStart = m_iVel_.linear.x;
    }
    speedPrev = speed;
    speedStart += speedAdd;
    if (speedAdd > 0) {
        if (speedStart > speed) {
            twist.linear.x = speed;
        } else {
            twist.linear.x = speedStart;
        }
    } else {
        if (speedStart < speed) {
            twist.linear.x = speed;

        } else {
            twist.linear.x = speedStart;

        }
    }

    if (twist.linear.x > m_dMaxLineSpeed_) {
        twist.linear.x = m_dMaxLineSpeed_;
    } else if (twist.linear.x < -m_dMaxLineSpeed_) {
        twist.linear.x = -m_dMaxLineSpeed_;
    }

    twist.angular.z = m_dDirP_ * m_iMagLineData_.bias + m_dDirD_ * (m_iMagLineData_.bias - last_error);
    last_error = m_iMagLineData_.bias;
    return false;
}

/**
 * @Title        :
 * @MethodName   :  Rotation
 * @param  in      :  angularSpeed （设置,原地旋转速度）
 * @param  out      :  twist (引用返回 速度)
 * @Return       :  bool  （true 表示旋转完成）
 * @Exception    :
 * @Description  :  原地旋转
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午1:47
 */
bool CMoveAction::Rotation(double angularSpeed, geometry_msgs::Twist &twist) {
    twist = geometry_msgs::Twist();
    switch (m_eTurnAroundStep_) {
        case e_STEP0:
            m_eTurnAroundStep_ = e_STEP1;
            break;
        case e_STEP1:
            twist.linear.x = 0;
            twist.angular.z = angularSpeed;
            if (m_iMagLineData_.lines == 0) {
                m_eTurnAroundStep_ = e_STEP2;
            }
            break;
        case e_STEP2:
            twist.linear.x = 0;
            twist.angular.z = angularSpeed;
            if (m_iMagLineData_.lines != 0) {
                m_eTurnAroundStep_ = e_STEP3;
            }
            break;
        case e_STEP3:
//            twist = tracingPd(0);
            twist.linear.x = 0;
            twist.angular.z = angularSpeed;
            if (std::abs(m_iMagLineData_.bias) < 5) {
                m_eTurnAroundStep_ = e_STEP_END;
            }
            break;
        case e_STEP_END:
            m_eTurnAroundStep_ = e_STEP0;
            return true;
            break;
    }
    if (twist.angular.z > m_dMaxAngularSpeed_) {
        twist.angular.z = m_dMaxAngularSpeed_;
    } else if (twist.angular.z < -m_dMaxAngularSpeed_) {
        twist.angular.z = -m_dMaxAngularSpeed_;
    }
    return false;
}

/**
 * @Title        :
 * @MethodName   :  Clear
 * @param null
 * @Return       :  null
 * @Exception    :
 * @Description  :  清除变量
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午1:49
 */
void CMoveAction::Clear() {
    m_eTurnAroundStep_ = e_STEP0;  // 原地旋转标志位清除
    m_eMoveMode_ = e_MOVE_IDLE;
    m_bChangeSpeeding_ = false;  // 方便保存速度平滑初始数据
    m_dStopDistance_ = 0;
}

/**
 * @Title        :
 * @MethodName   :  VelCallback
 * @param in     :  msg
 * @Return       :  null
 * @Exception    :
 * @Description  :  订阅机器人，轮子速度
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午2:44
 */
void CMoveAction::VelCallback(const nav_msgs::Odometry::ConstPtr &msg) {
    m_iVel_ = msg->twist.twist;
    static double timePrev = ros::Time::now().toSec();
    if(m_bIsStop_){
        m_dStopDistance_ += m_iVel_.linear.x * ( ros::Time::now().toSec() - timePrev);
    }else{
        m_dStopDistance_ = 0;
    }
    timePrev =  ros::Time::now().toSec();
}

/**
 * @Title        :
 * @MethodName   :  advance
 * @param     in :  speed
 * @param    out :  twist
 * @Return       :  null
 * @Exception    :
 * @Description  :  机器人巡线向前，如果前面没有磁条则旋转一圈继续巡线
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午2:57
 */
bool CMoveAction::Advance(double speed, geometry_msgs::Twist &twist,bool useDistance) {

    if(useDistance){
        static double errorLast{0};
        m_bIsStop_ = true;
        double stopDistance = speed;
        double error = stopDistance - m_dStopDistance_;
        speed  =  m_dSpeedP_ * error + m_dSpeedD_ * (error - errorLast);
        if(speed < 0.05){
            speed = 0.05;
        }
    } else{
        m_bIsStop_ = false;
    }
    if (m_iMagLineData_.lines > 0 and m_eTurnAroundStep_ == e_STEP0) {
        TracingPd(speed, twist);
    } else {
        if (Rotation(m_dAngularSpeed_, twist)) {
            m_eTurnAroundStep_ = e_STEP0;
        }
    }
    return false;
}

/**
 * @Title        :
 * @MethodName   :  advanceStop
 * @param in     :  speed 设置速度
 * @param out    :  twist 反馈速度
 * @Return       :
 * @Exception    :
 * @Description  :  巡线，没有磁条的时候，停止
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午3:13
 */
bool CMoveAction::AdvanceStop(double speed, geometry_msgs::Twist &twist) {
    if (m_iMagLineData_.lines > 0) {
        TracingPd(speed, twist);
    } else {
        twist = geometry_msgs::Twist();
    }
    return false;
}

/**
 * @Title        :
 * @MethodName   :  stopQuick
 * @param out    :  twist
 * @Return       :
 * @Exception    :
 * @Description  :  快速停止，机器人可能会发生振动
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午3:08
 */
bool CMoveAction::StopQuick(geometry_msgs::Twist &twist) {
    twist = geometry_msgs::Twist();
    return false;
}

/**
 * @Title        :
 * @MethodName   :  stopSmooth
 * @param out    :  twist
 * @Return       :  bool （ture 表示停止成功）
 * @Exception    :
 * @Description  :  平滑停止
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午3:10
 */
bool CMoveAction::StopSmooth(geometry_msgs::Twist &twist) {
    if (std::abs(m_iVel_.linear.x) < 0.01) {
        twist.linear.x = 0;
        return true;
    } else {
        twist.linear.x = m_iVel_.linear.x - 0.02;
    }
    return false;
}

/**
 * @Title        :
 * @MethodName   :  SetMode
 * @param in     :  mode
 * @Return       :
 * @Exception    :
 * @Description  :  设置运动模式，前进，后退还是停止
 *
 * @author       :  王朋朋
 * @date         :  2021/4/23 下午1:21
 */
void CMoveAction::SetMode(MOVE_MODE mode) {
    m_eMoveMode_ = mode;
}

/**
 * @Title        :
 * @MethodName   :  ChangeSpeedWithDistance
 * @param in     :  speed 设定速度
 * @param in     :  distance 设定距离
 * @param out    :  twist 反馈速度
 * @Return       :
 * @Exception    :
 * @Description  :  根据距离，修改速度达到设定值。主要功能速度平滑
 *
 * @author       :  王朋朋
 * @date         :  2021/4/23 下午1:31
 */
bool CMoveAction::ChangeSpeedWithDistance(double speed, double distance, geometry_msgs::Twist &twist) {
    static double startSpeed{0};
    static double lastSpeed{0};
    static double pastDistance{0};
    static double lastTime{0};
    static double a{0};
    static double b{0};   // 直线方程 y = ax + b

    if (m_bChangeSpeeding_) { // 第一次运行修改速度，保持初始速度信息，让步计算差值。公式参数
        startSpeed = m_iVel_.linear.x;
        m_bChangeSpeeding_ = true;
        lastTime = ros::Time::now().toSec();
        lastSpeed = m_iVel_.linear.x;
        a = (speed - startSpeed) / distance;
        b = startSpeed;

    } else {
        pastDistance += lastSpeed * (ros::Time::now().toSec() - lastTime);
        twist.linear.x = a * pastDistance + b;
    }
    if (std::abs(twist.linear.x - speed) < 0.03) {
        m_bChangeSpeeding_ = false;
        return true;
    }
    return false;
}

/**
 * @Title        :
 * @MethodName   :  SetRotationStep
 * @param in     :  moveStep
 * @Return       :  null
 * @Exception    :
 * @Description  :  设置旋转步骤，主要防止上次旋转没结束就重新开始
 *
 * @author       :  王朋朋
 * @date         :  2021/4/25 下午2:38
 */
bool CMoveAction::SetRotationStep(MOVE_STEP moveStep) {
    m_eTurnAroundStep_ = moveStep;
    return false;
}

/**
 * @Title        :
 * @MethodName   :  InitParam
 * @param null
 * @Return       :
 * @Exception    :
 * @Description  :  初始化对象相关参数
 *
 * @author       :  王朋朋
 * @date         :  2021/4/25 下午3:47
 */
void CMoveAction::InitParam() {
    MyGetParam("/move_action_param/speed_p", m_dSpeedP_, 1.0);
    MyGetParam("/move_action_param/speed_d", m_dSpeedD_, 1.0);
    MyGetParam("/move_action_param/dir_p", m_dDirP_, 1.0);
    MyGetParam("/move_action_param/dir_d", m_dDirD_, 1.0);
    MyGetParam("/move_action_param/max_line_speed", m_dMaxLineSpeed_, 1.0);
    MyGetParam("/move_action_param/max_angular_speed", m_dMaxAngularSpeed_, 1.0);
    MyGetParam("/move_action_param/speed_angular", m_dAngularSpeed_, 1.0);
    Clear();
}

/**
 * @Title        :
 * @MethodName   :  MyGetParam
 * @param in     :  paramName
 * @param in     :  defaultValue
 * @param out    :  param
 * @Return       :  null
 * @Exception    :
 * @Description  :  判断参数服务器是否包含相关参数，不包含则输出告警信息。使用默认数据
 *
 * @author       :  王朋朋
 * @date         :  2021/4/25 下午3:48
 */
template<typename T>
void CMoveAction::MyGetParam(std::string paramName, T &param, T defaultValue) {
    if (not m_iNh_.hasParam(paramName))
        ROS_WARN("[警告] 未找到参数 ： %s , 使用默认参数 %s", paramName.data(), std::to_string(defaultValue).data());
    m_iNh_.param<T>(paramName, param, defaultValue);
}
