/*
 * @Descripttion: navigation.cpp Project
 * @version: Project v0.1
 * @Author: lenotary
 * @Date: 2021/4/21 下午2:07
 * @LastEditor: lenotary
 * @LastEditTime: 2021/4/21 下午2:07
 */




#include "navigation.h"

/**
 * @Title        :
 * @MethodName   :  CNavigation
 * @param in     :  nh ros node 节点句柄
 * @Return       :
 * @Exception    :
 * @Description  :  初始化 action server
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午3:45
 */
CNavigation::CNavigation(ros::NodeHandle &nh) : m_iNh_(nh), m_iActionServer_(nh, "/robot_pos_ctrl", [this](auto &&PH1) {
    ExecuteCb(PH1);
}, false) {
    ROS_INFO("初始化 Navigation 对象 ... ");
    m_iJoySpeed_ = geometry_msgs::Twist();
    m_iActionServer_.registerPreemptCallback([this] { PreemptCb(); });
    m_iRfidSub_ = nh.subscribe("/device/rfid_data", 1, &CNavigation::RfidCallback, this);
    m_iCoproSwitchSub_ = nh.subscribe("/copro_switch", 10, &CNavigation::CoproSwitchCallback, this);
    m_iSpecialControlSub_ = nh.subscribe("/task_special_ctrl", 10, &CNavigation::SpecialControlCallback, this);
    m_iJoySub_ = nh.subscribe("/label_vel", 10, &CNavigation::LabelVelCallback, this);
    m_iModeSub_ = nh.subscribe("/sys/mode_switch", 10, &CNavigation::ModeCallback, this);

    m_iVelPub_ = nh.advertise<geometry_msgs::Twist>("/cmd_vel", 1);
    m_iMagLineSelectPub_ = nh.advertise<user_msgs::MagLineSelect>("/device/mag_line_select", 1);
    m_iRobotPosePub_ = nh.advertise<user_msgs::robot_pos>("/robot_pos", 1);
    m_pfnMoveAction_ = new CMoveAction(nh);
    InitParam();
    m_iActionServer_.start();
    ROS_INFO("初始化成功");
}

/**
 * @Title        :
 * @MethodName   :  ExecuteCb
 * @param in     :  goal
 * @Return       :  null
 * @Exception    :
 * @Description  :  action 服务回调函数，当有任务从客户端发过来的时候，就会调用这个函数。任务结束，则跳出这个函数
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午3:45
 *
 *
 *
 */
void CNavigation::ExecuteCb(const user_msgs::robot_pos_ctrlGoalConstPtr &goal) {
    m_bActionActive = true;
    user_msgs::robot_pos_ctrlResult result;
    Clear();
    ROS_INFO("[导航任务开始] ++++++++++++++++++++++++++++++");
    ROS_INFO("接受到新任务, 路径大小 %ld", goal->path.size());
    m_vPath_ = goal->path;
    //打印路径点信息，把ID保存成列表，方便后面判断点是否在路径里
    for (int i = 0; i < m_vPath_.size(); i++) {
        auto point = goal->path[i];
        m_vIdList_.push_back(static_cast<int>(point.X));
        ROS_INFO("点位信息 Index : %d , ID：%d , 方向 %d , 类型 %d", i, static_cast<int>(point.X), static_cast<int>(point.yaw),
                 point.type);
    }
    // 计算终点在路径中出现的次数，主要为了有的路径终点出现两次情况，比如需要掉头回到终点。或者循环多次任务等
    for (int i = 1; i < m_vPath_.size(); ++i) {
        if (m_vPath_[i].X == m_vPath_.back().X) {
            m_nGoalTime_++;
        }
    }
    m_iLineSelect_.select = static_cast<int>(m_vPath_[m_nIdIndex_].yaw);
    ROS_INFO("[信息] 选择路线 %d %f", m_iLineSelect_.select,m_vPath_[m_nIdIndex_].X);
    ROS_INFO("终点需要经过的次数 %d", m_nGoalTime_);
//    for (auto point: m_vPath_) {
//        if (point.X == m_vPath_.back().X) {
//            m_nGoalTime_++;
//        }
//    }
    geometry_msgs::Twist twist;
    user_msgs::robot_pos_ctrlFeedback feedback;
    feedback.drived = 1;
    ros::Rate rate(50);
    while (m_bActionActive and ros::ok()) {
        switch (m_eActionState) {
            case e_ACTION_IDLE:
                ROS_INFO("[信息] 开始导航");
                m_eActionState = e_ACTION_RUNNING;
                break;
            case e_ACTION_RUNNING:
                m_pfnMoveAction_->Advance(m_dNavSpeedMid_, twist);
                break;
            case e_ACTION_PARKING:

                m_pfnMoveAction_->Advance(0.15, twist, true);
                break;
            case e_ACTION_ROTATION:
                if (m_pfnMoveAction_->Rotation(m_dRotationSpeedMax_, twist)) {
                    m_eActionState = e_ACTION_RUNNING;
                    ROS_INFO("[信息] 机器人原地旋转结束");
                }
                break;
            case e_ACTION_END:
                m_pfnMoveAction_->StopQuick(twist);
                m_eActionState = e_ACTION_IDLE;
                result.res = "ok";
                m_iActionServer_.setSucceeded(result);
                ROS_INFO("[导航]  导航任务完成 ");
                m_bActionActive = false; //任务完成
                break;
            case e_ACTION_POINT_ERROR:
                if (m_pfnMoveAction_->StopSmooth(twist)) {
                    result.res = "fail";
                    m_iActionServer_.setSucceeded(result);
                    ROS_ERROR("[错误] 机器人停止");
                    m_eActionState = e_ACTION_IDLE;
                    m_bActionActive = false; // 任务完成；
                }
                break;
        }
        if(m_bIsObstacle_){
            twist = geometry_msgs::Twist();
        }
        m_iActionServer_.publishFeedback(feedback);
        m_iMagLineSelectPub_.publish(m_iLineSelect_);
        m_iVelPub_.publish(twist);
        rate.sleep();
    }
    ROS_INFO("[导航任务结束] ------------------------------");
}

/**
 * @Title        :
 * @MethodName   :  PreemptCb
 * @param null
 * @Return       :
 * @Exception    :
 * @Description  :  任务执行中，如果任务被中断，则任务停止清除标志为
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午3:47
 */
void CNavigation::PreemptCb() {
    ROS_WARN("任务被中断");
    geometry_msgs::Twist twist;
    while (!m_pfnMoveAction_->StopSmooth(twist)) {
        ros::Duration(0.05).sleep();
        m_iVelPub_.publish(twist);
    }
    m_iVelPub_.publish(twist);
    m_pfnMoveAction_->Clear();
    m_bActionActive = false;
    user_msgs::robot_pos_ctrlResult result;
    result.res = "fail";
    m_iActionServer_.setSucceeded(result);
    ROS_WARN("机器人停止");


}

/**
 * @Title        :
 * @MethodName   :  RfidCallback
 * @param in     :  msg
 * @Return       :
 * @Exception    :
 * @Description  :  订阅 RFID信息，逻辑处理，根据当前RFID的信息，判断点的类型和下一个目标点需要行走的方向。
 *                  判断是否需要在本次目标点停下。还是旋转。
 *
 *
 *
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午5:17
 */
void CNavigation::RfidCallback(const user_msgs::RfidMsg_<std::allocator<void>>::ConstPtr &msg) {
    static int lastId{0};
    ROS_WARN("读取到RFID信息 : %d", msg->id);
    user_msgs::robot_pos robotPosMsg;
    robotPosMsg.x = msg->id*1000;
    m_iRobotPosePub_.publish(robotPosMsg);
    if (m_eActionState == e_ACTION_IDLE)
        return;
    lastId = m_nCurrentId_;
    m_nCurrentId_ = msg->id;
    
    m_vPastId_.push_back(m_nCurrentId_);
    ROS_INFO("点位索引 : %d", m_nIdIndex_);
    ROS_INFO("经过的路径有 : %s", VectorToStr(m_vPastId_).data());
    if(GetPointType(m_nCurrentId_) == e_ROTATION_POINT and lastId == m_nCurrentId_){
        ROS_INFO("原地旋转");
        return;
    }
    if (m_nIdIndex_ > m_vPath_.size() - 1) {
        ROS_INFO("[错误] 路径点错误，路径过多。机器人停止");
        m_eActionState = e_ACTION_POINT_ERROR;
        return;
    }
    if (m_nCurrentId_ != m_vPath_[m_nIdIndex_].X) {
        ROS_INFO("[错误] 路径点错误，错误点。机器人停止 %d %f", m_nCurrentId_, m_vPath_[m_nIdIndex_].X);
        m_eActionState = e_ACTION_POINT_ERROR;
        return;
    }
    m_nIdIndex_ += 1;
    if(m_nIdIndex_ < m_vPath_.size()){
        m_iLineSelect_.select = static_cast<int>(m_vPath_[m_nIdIndex_].yaw);
        ROS_INFO("[信息] 选择路线 %d %f", m_iLineSelect_.select,m_vPath_[m_nIdIndex_].X);
    }else{
        ROS_INFO("最后一个点");
    }
    if (not IsInclude(m_vIdList_, m_nCurrentId_)) {
        m_eActionState = e_ACTION_POINT_ERROR;
        ROS_INFO("[错误] 路径点错误，路径不包含当前点。机器人停止");
    }
    //判断旋转点是否是第一次经过，由于机器人旋转过后会再经过一次旋转点。所以第二次不会旋转。默认机器人只会经过一次旋转点
    if (GetPointType(m_nCurrentId_) == e_ROTATION_POINT) {
        m_eActionState = e_ACTION_ROTATION;
        ROS_INFO("到达旋转点，减速准备原地旋转");
    } else if (m_nCurrentId_ == static_cast<int>(m_vPath_.back().X) or static_cast<int>(m_vPath_.back().X) == -1) {
        int pastGoalTimes{0}; //计算机器经过终点次数。
        for (auto id : m_vPastId_) {
            if (id == m_vPath_.back().X) {
                pastGoalTimes++;
            }
        }
        ROS_INFO("机器人经过终点次数 ： %d", pastGoalTimes);
        if (pastGoalTimes == m_nGoalTime_) {
            ROS_INFO("到达终点,机器人减速停止");
            m_eActionState = e_ACTION_PARKING;
        }
    }


}

/**
 * @Title        :
 * @MethodName   :  CoproSwitchCallback
 * @param in     :  msg
 * @Return       :  null
 * @Exception    :
 * @Description  :  订阅激光测距传感器，主要是检测下面RFID反光板，检测到反光板表示到达目标点需要立即停止。
 *
 * @author       :  王朋朋
 * @date         :  2021/4/25 上午10:14
 */
void CNavigation::CoproSwitchCallback(const user_msgs::CoproSwitch::ConstPtr &msg) {
    if(msg->cmd == user_msgs::CoproSwitch::CMD_PHOTO_SWITCH){
        if (m_eActionState == e_ACTION_PARKING) {
            if (GetPointType(m_nCurrentId_) == e_ROTATION_POINT) {
                m_eActionState = e_ACTION_ROTATION;
                m_pfnMoveAction_->SetRotationStep(e_STEP0);
                ROS_INFO("开始原地旋转");
            } else if (GetPointType(m_nCurrentId_) == e_STOP_POINT or GetPointType(m_nCurrentId_) == e_START_POINT or GetPointType(m_nCurrentId_) == e_CHARGE_POINT) {
                m_eActionState = e_ACTION_END;
                ROS_INFO("机器人到达目标点，停止");
            }
        }
        m_bIsObstacle_ = false;
    }else if(msg->cmd == user_msgs::CoproSwitch::CMD_OBSTACLE_LASER){
        if(msg->param == 1){
            m_bIsObstacle_ = true;
            ROS_WARN("[警告] 前方有障碍物");
        }else{
            m_bIsObstacle_ = false;
            ROS_WARN("[警告] 障碍物消失");
        }

    }else{
        m_bIsObstacle_ = false;
    }

}
void CNavigation::SpecialControlCallback(const user_msgs::special_ctrl::ConstPtr &msg) {
    if(msg->stop){
        ROS_WARN("[警告]紧急停止");
        PreemptCb();
    }
    if(msg->release){
        ROS_WARN("[警告]取消急停");
    }
}
void CNavigation::LabelVelCallback(const geometry_msgs::Twist::ConstPtr &msg) {
    if(!m_bActionActive){
        geometry_msgs::Twist twist;
        if(msg->angular.z > 0.05){
            m_iLineSelect_.select = 1;
        }else if(msg->angular.z < -0.05){
            m_iLineSelect_.select = 3;
        } else{
            m_iLineSelect_.select = 2;
        }
        if(std::abs(msg->linear.x) > 0.05){
            m_pfnMoveAction_->Advance(msg->linear.x,twist);
        }else{
            twist.angular.z = msg->angular.z;
        }
        m_iMagLineSelectPub_.publish(m_iLineSelect_);
        m_iVelPub_.publish(twist);








    }

}void CNavigation::ModeCallback(const user_msgs::SysModeSwitch::ConstPtr &msg) {
    static user_msgs::SysModeSwitch sysModeSwitchPrev;
    if(msg->mode != sysModeSwitchPrev.mode){
        ROS_INFO("机器人模式 %d",msg->mode);
    }
    m_iMode_ = *msg;
}
/**
 * @Title        :
 * @MethodName   :  Clear
 * @param null
 * @Return       :
 * @Exception    :
 * @Description  :  清除导航需要的相关信息和标志位，每次导航结束的时候清除一下。
 *
 * @author       :  王朋朋
 * @date         :  2021/4/22 下午5:21
 */
void CNavigation::Clear() {
    ROS_INFO("清除上次Action, 相关缓存");
    m_vPastId_.clear();
    m_vPath_.clear();
    m_pfnMoveAction_->Clear();
    m_eActionState = e_ACTION_IDLE;
    m_nIdIndex_ = 1;
    m_nGoalTime_ = 0;
}

/**
 * @Title        :
 * @MethodName   :  VectorToStr
 * @param in     :  list
 * @Return       :
 * @Exception    :
 * @Description  :  把数组内数据变成字符串，方便打印
 *
 * @author       :  王朋朋
 * @date         :  2021/4/23 上午11:38
 */

std::string CNavigation::VectorToStr(std::vector<int> list) {
    std::string str;

    for (auto data:list) {
        str += "[" + std::to_string(data) + "] ";
    }
    return str;
}

/**
 * @Title        :
 * @MethodName   :  IsInclude
 * @param in     :  list 数组
 * @param in     :  key     需要判断的数据
 * @Return       :  true 表示数组内保护 相应元素
 * @Exception    :
 * @Description  :  判断数组是否包含某元素
 *
 * @author       :  王朋朋
 * @date         :  2021/4/23 上午11:39
 */
bool CNavigation::IsInclude(std::vector<int> list, int key) {
    if (std::find(list.begin(), list.end(), key) != list.end()) {
        return true;
    }
    return false;
}

/**
 * @Title        :
 * @MethodName   :  GetPointType
 * @param in     :  id
 * @Return       :  pointType
 * @Exception    :
 * @Description  :  根据id号，检索相应的点的类型
 *
 * @author       :  王朋朋
 * @date         :  2021/4/23 上午11:42
 */
int CNavigation::GetPointType(int id) {

    for (auto point : m_vPath_) {
        if (static_cast<int>(point.X) == id) {
            return point.type;
        }
    }
    return e_OTHER_POINT;
}

/**
 * @Title        :
 * @MethodName   :  InitParam
 * @param null
 * @Return       :
 * @Exception    :
 * @Description  :  初始化对象相关参数
 *
 * @author       :  王朋朋
 * @date         :  2021/4/25 下午3:47
 */
void CNavigation::InitParam() {
    MyGetParam("/nav_param/nav_speed_max", m_dNavSpeedMax_, 0.8);
    MyGetParam("/nav_param/nav_speed_mid", m_dNavSpeedMid_, 0.5);
    MyGetParam("/nav_param/nav_speed_min", m_dNavSpeedMin_, 0.2);
    MyGetParam("/nav_param/speed_angular", m_dRotationSpeedMax_, 0.2);
    m_iLineSelect_.select = 1;
    m_eActionState = e_ACTION_IDLE;
}

/**
 * @Title        :
 * @MethodName   :  MyGetParam
 * @param in     :  paramName
 *
 *
 * @param in     :  defaultValue
 * @param out    :  param
 * @Return       :  null
 * @Exception    :
 * @Description  :  判断参数服务器是否包含相关参数，不包含则输出告警信息。使用默认数据
 *
 * @author       :  王朋朋
 * @date         :  2021/4/25 下午3:48
 */
template<typename T>
void CNavigation::MyGetParam(std::string paramName, T &param, T defaultValue) {
    if (not m_iNh_.hasParam(paramName))
        ROS_WARN("[警告] 未找到参数 ： %s , 使用默认参数 %s", paramName.data(), std::to_string(defaultValue).data());
    m_iNh_.param<T>(paramName, param, defaultValue);
}




